import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';
import { GUI } from './jsm/libs/dat.gui.module.js';
import { OrbitControls } from './jsm/controls/OrbitControls.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        let scene = new THREE.Scene();
        scene.background = new THREE.Color(0xa0a0a0);
        //scene.fog = new THREE.Fog(0xa0a0a0, 10, 22);

        this.Empty = Object.freeze([]);

        let renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.shadowMap.enabled = true;
        renderer.clippingPlanes = this.Empty;
        renderer.localClippingEnabled = true;
        this.domWebGL.appendChild(renderer.domElement);

        this.renderer = renderer;
        this.scene = scene;

        this._stats = new Stats();
        this.domWebGL.appendChild(this._stats.dom);

        this._setupCamera();
        this._setupModel();
        this._setupLights();
        this._setupControls();
        this._setupGUI();

        window.onresize = this.resize.bind(this);

        this.resize();
    }

    _setupControls() {
        const controls = new OrbitControls(this._camera, this.renderer.domElement);
        controls.minDistance = 1;
        controls.maxDistance = 100;
        controls.target.set(0, 1, 0);
        controls.update();
    }

    _setupGUI() {
        const that = this;
        const globalClippingPlanes = this._createPlanes(this._GlobalClippingPlanes.length);

        const gui = new GUI(),
            folder = gui.addFolder("Local Clipping"),
            props = {
                get Enabled() {
                    return that.renderer.localClippingEnabled;
                },
                set Enabled(v) {
                    that.renderer.localClippingEnabled = v;
                    if(!v) that._volumnVisualization.visible = false;
                },
                get Shadows() {
                    return that._clipMaterial.clipShadows;
                },
                set Shadows(v) {
                    that._clipMaterial.clipShadows = v;
                },
                get Visualize() {
                    return that._volumnVisualization.visible;
                },
                set Visualize(v) {
                    if(that.renderer.localClippingEnabled) {
                        that._volumnVisualization.visible = v;
                    }
                }
            };

        folder.add(props, "Enabled");
        folder.add(props, "Shadows");
        folder.add(props, "Visualize").listen();

        gui.addFolder("Global Clipping").
            add({
                get "Enabled"() {
                    return that.renderer.clippingPlanes !== that.Empty;
                },
                set "Enabled"(v) {
                    that.renderer.clippingPlanes = v ? globalClippingPlanes : that.Empty;
                }
            }, "Enabled");

        this._startTime = Date.now();
        this._globalClippingPlanes = globalClippingPlanes;
    }

    _planesFromMesh(vertices, indices) {
        const n = indices.length / 3, result = new Array(n);

        for(let i=0, j=0; i<n; ++i, j+=3) {
            const a = vertices[indices[j]],
                b = vertices[indices[j+1]],
                c = vertices[indices[j+2]];

            result[i] = new THREE.Plane().setFromCoplanarPoints(a, b, c);
        }

        return result;
    }

    _createPlanes(n) {
        const result = new Array(n);

        for(let i=0; i !== n; i++) {
            result[i] = new THREE.Plane();
        }

        return result;
    }

    _cylindricalPlanes(n, innerRadius) {
        const result = this._createPlanes(n);

        for(let i=0; i!==n; i++) {
            const plane = result[i], angle = i*Math.PI*2/n;
        
            plane.normal.set(Math.cos(angle), 0, Math.sin(angle));
            plane.constant = innerRadius;
        }
        return result;
    }

    // creates a matrix that aligns X/Y to a given plane
    _planeToMatrix(plane) {
        const xAxis = new THREE.Vector3(), 
            yAxis = new THREE.Vector3(), trans = new THREE.Vector3(),
            zAxis = plane.normal, matrix = new THREE.Matrix4();

        if (Math.abs(zAxis.x) > Math.abs(zAxis.z)) {
            yAxis.set(-zAxis.y, zAxis.x, 0);
        } else {
            yAxis.set(0, -zAxis.z, zAxis.y);
        }

        xAxis.crossVectors(yAxis.normalize(), zAxis);
        plane.coplanarPoint(trans);

        return matrix.set(
            xAxis.x, yAxis.x, zAxis.x, trans.x,
            xAxis.y, yAxis.y, zAxis.y, trans.y,
            xAxis.z, yAxis.z, zAxis.z, trans.z,
            0, 0, 0, 1);
    }

    _setupModel() {
        const Vertices = [
                new THREE.Vector3(1, 0, Math.SQRT1_2),
                new THREE.Vector3(-1, 0, Math.SQRT1_2),
                new THREE.Vector3(0, 1, -Math.SQRT1_2),
                new THREE.Vector3(0, -1, -Math.SQRT1_2)
            ],
            Indices = [
                0,1,2,
                0,2,3,
                0,3,1,
                1,3,2
            ],
            Planes = this._planesFromMesh(Vertices, Indices),
            PlaneMatrices = Planes.map(this._planeToMatrix),
            GlobalClippingPlanes = this._cylindricalPlanes(5, 2.5);
        
        const clipMaterial = new THREE.MeshPhongMaterial({
            color: 0xee0a10,
            shininess: 100,
            side: THREE.DoubleSide,
            clippingPlanes: this._createPlanes(Planes.length),
            clipShadows: true
        });

        const object = new THREE.Group();
        const geometry = new THREE.BoxGeometry(0.18, 0.18, 0.18);
        for(let z = -2; z <= 2; z++) {
            for(let y = -2; y <= 2; y++) {
                for(let x = -2; x <= 2; x++) {
                    const mesh = new THREE.Mesh(geometry, clipMaterial);
                    mesh.position.set(x/5, y/5, z/5);
                    mesh.castShadow = true;
                    object.add(mesh);
                }
            }
        }

        const planeGeometry = new THREE.PlaneGeometry(3, 3, 1, 1), 
            color = new THREE.Color();

        const volumnVisualization = new THREE.Group();
        volumnVisualization.visible = false;

        for(let i=0, n=Planes.length; i !== n; i++) {
            const material = new THREE.MeshBasicMaterial({
                color: color.setHSL(i/n, 0.5, 0.5).getHex(),
                side: THREE.DoubleSide,
                opacity: 0.2,
                transparent: true,
                clippingPlanes: clipMaterial.clippingPlanes.filter( (_, j) => {
                    return j!==i;
                })
            });

            const mesh = new THREE.Mesh(planeGeometry, material);
            mesh.matrixAutoUpdate = false;
            volumnVisualization.add(mesh);
        }

        this.scene.add(volumnVisualization);

        const ground = new THREE.Mesh(planeGeometry,
            new THREE.MeshPhongMaterial({ color: 0xa0adaf, shininess: 10}));
        ground.rotation.x = -Math.PI / 2;
        ground.scale.multiplyScalar(3);
        ground.receiveShadow = true;
        this.scene.add(ground);

        this.scene.add(object);

        this._volumnVisualization = volumnVisualization;
        this._object = object;
        this._clipMaterial = clipMaterial;
        this._Planes = Planes;
        this._PlaneMatrices = PlaneMatrices;
        this._GlobalClippingPlanes = GlobalClippingPlanes;
    }

    _setupLights() {
        this.scene.add(new THREE.AmbientLight(0xffffff, 0.3));

        const spotLight = new THREE.SpotLight(0xffffff, 0.5);
        spotLight.angle = Math.PI / 5;
        spotLight.penumbra = 0.2;
        spotLight.position.set(2,3,3);
        spotLight.castShadow = true;
        /*
        spotLight.shadow.camera.near = 3;
        spotLight.shadow.camera.far = 10;
        spotLight.shadow.mapSize.width = 1024;
        spotLight.shadow.mapSize.height = 1024;
        */
        this.scene.add(spotLight);

        const dirLight = new THREE.DirectionalLight(0xffffff, 0.5);
        dirLight.position.set(0,2,0);
        dirLight.castShadow = true;
        /*
        dirLight.shadow.camera.near = 1;
        dirLight.shadow.camera.far = 10;
        dirLight.shadow.camera.right = 1;
        dirLight.shadow.camera.left = -1;
        dirLight.shadow.camera.top = 1;
        dirLight.shadow.camera.bottom = -1;
        dirLight.shadow.mapSize.width = 1024;
        dirLight.shadow.mapSize.height = 1024;
        */
        this.scene.add(dirLight);


    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(36, window.innerWidth / window.innerHeight, 0.25, 50);
        camera.position.set(0, 1.5, 15);

        //this.scene.add(camera);

        this._camera = camera;
    }

    _assignTransformedPlanes(planesOut, planesIn, matrix) {
        for(let i=0, n=planesIn.length; i !== n; ++i) {
            planesOut[i].copy(planesIn[i]).applyMatrix4(matrix);
        }
    }

    _setObjectWorldMatrix(object, matrix) {
        const parent = object.parent;
        this.scene.updateMatrixWorld();
        object.matrix.copy(parent.matrixWorld).invert();
        object.applyMatrix4(matrix);
    }

    update() {
        const currentTime = Date.now(), time = (currentTime - this._startTime) / 1000;

        this._object.position.y = 1;
        this._object.rotation.x = time * 0.5;
        this._object.rotation.y = time * 0.2;
        this._object.updateMatrix();

        const transform = new THREE.Matrix4(), tmpMatrix = new THREE.Matrix4();

        transform.copy(this._object.matrix);
        const bouncy = Math.cos(time * 0.5) * 0.5 + 0.7;
        transform.multiply(tmpMatrix.makeScale(bouncy, bouncy, bouncy));

        this._assignTransformedPlanes(this._clipMaterial.clippingPlanes, this._Planes, transform);
        const planeMeshes = this._volumnVisualization.children;
        for(let i=0, n=planeMeshes.length; i !== n; i++) {
            tmpMatrix.multiplyMatrices(transform, this._PlaneMatrices[i]);
            this._setObjectWorldMatrix(planeMeshes[i], tmpMatrix);
        }

        transform.makeRotationY(time * 0.1);
        this._assignTransformedPlanes(this._globalClippingPlanes, this._GlobalClippingPlanes, transform);
    }

    render(time) {
        requestAnimationFrame(this.render.bind(this));

        this.update(time);
        
        this._stats.begin();
        this.renderer.render(this.scene, this._camera);
        this._stats.end();
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setSize(window.innerWidth, window.innerHeight);

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        this.render();
    }
}

window.onload = function () {
    (new App()).render();
}